from q2gui.pyqt6.widgets import (
    q2label,
    q2image,
    q2line,
    q2frame,
    q2check,
    q2toolbar,
    q2tab,
    q2text,
    q2code,
    q2button,
    q2relation,
    q2radio,
    q2lookup,
    q2combo,
    q2toolbutton,
    q2progressbar,
    q2grid,
    q2sheet,
    q2scroller,
    q2date,
    q2space,
    q2list,
    q2spin,
    q2doublespin,
    q2time,
    q2color,
)


(
    q2label,
    q2image,
    q2line,
    q2frame,
    q2check,
    q2toolbar,
    q2tab,
    q2text,
    q2code,
    q2button,
    q2relation,
    q2radio,
    q2lookup,
    q2combo,
    q2toolbutton,
    q2progressbar,
    q2grid,
    q2sheet,
    q2scroller,
    q2date,
    q2space,
    q2list,
    q2spin,
    q2doublespin,
    q2time,
    q2color,
)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import QComboBox, QApplication
from PyQt6.QtCore import Qt, QEvent
from PyQt6.QtGui import QKeyEvent
from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.q2utils import int_


class q2combo(QComboBox, Q2Widget):
    def __init__(self, meta):
        super().__init__(meta)
        self.meta = meta
        self.set_data(meta.get("pic", ""))
        if self.meta.get("data"):
            self.set_text(self.meta.get("data"))
        else:
            self.setCurrentIndex(0)
        self.currentIndexChanged.connect(self.valid)

    def set_data(self, data):
        self.clear()
        if isinstance(data, str):
            data = data.split(";")
        width = 0
        for item in data:
            self.addItem(item)
            width = max(len(item), width)
        if int_(self.meta.get("datalen", 0)) != 0:
            width = int_(self.meta.get("datalen", 0))
        self.set_minimum_width(width)

    def keyPressEvent(self, ev):
        if ev.key() in [
            Qt.Key.Key_Down,
            Qt.Key.Key_Up,
            Qt.Key.Key_PageDown,
            Qt.Key.Key_PageUp,
        ]:
            ev.ignore()
        elif ev.key() == Qt.Key.Key_Right:
            QApplication.sendEvent(self, QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Tab, ev.modifiers()))
        elif ev.key() == Qt.Key.Key_Left:
            QApplication.sendEvent(
                self, QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Tab, Qt.KeyboardModifier.ShiftModifier)
            )
        else:
            super().keyPressEvent(ev)

    def set_text(self, text):
        if self.meta.get("num") or isinstance(text, int):
            index = int_(text)
            index = index - 1 if index else 0
        else:
            index_list = [x for x in range(self.count()) if self.itemText(x) == text]
            if index_list:
                index = index_list[0]
            else:
                index = 0
        self.setCurrentIndex(index)

    def get_text(self):
        if self.currentText():
            if self.meta.get("num"):
                return self.currentIndex() + 1
            else:
                return self.currentText()
        else:
            return ""

    def set_readonly(self, arg):
        self.setDisabled(arg)
        return super().set_readonly(arg)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import QDialog, QMdiSubWindow, QApplication, QTabBar
from PyQt6.QtCore import Qt, QEvent
from PyQt6.QtGui import QKeySequence, QKeyEvent


import q2gui.q2app as q2app
import q2gui.q2form as q2form
import q2gui.pyqt6.widgets

from q2gui.pyqt6.q2app import Q2QtWindow
from q2gui.q2utils import num

import q2gui.q2dialogs

# from q2gui.q2dialogs import q2Mess, q2Wait, q2AskYN

from q2gui.pyqt6.q2widget import Q2Widget


class Q2Form(q2form.Q2Form):
    def __init__(self, title=""):
        super().__init__(title=title)
        self._Q2FormWindow_class = Q2FormWindow
        self._q2dialogs = q2gui.q2dialogs
        # if QApplication.activeWindow():
        #     self.q2_app = QApplication.activeWindow()
        # else:
        #     self.q2_app = q2app.q2_app
        self.q2_app = q2app.q2_app
        self.on_init()

    def grid_navigation_actions_hook(self, actions):
        pass


class Q2FormWindow(QDialog, q2form.Q2FormWindow, Q2QtWindow, Q2Widget):
    def __init__(self, q2_form: Q2Form, title=""):
        super().__init__(q2_form, title)
        title = title if title else q2_form.title
        # Q2QtWindow.__init__(self, self.title)
        self.set_title(title)
        self._widgets_package = q2gui.pyqt6.widgets
        self.setObjectName("q2form")
        self.layout().setContentsMargins(2, 2, 2, 2)
        self.layout().setSpacing(0)
        self.heap = q2app.Q2Heap()
        self.heap.modal = None
        self.heap.prev_mdi_window = None
        self.heap.prev_focus_widget = None
        self.heap.prev_tabbar_text = ""
        self.about_to_close = False

    def restore_geometry(self, settings):
        paw = self.parent()

        if paw is not None:
            # save default == minimal size
            sizeBefore = paw.size()

            if self.q2_form.do_not_save_geometry:
                width, height = 1, 1
            else:
                width = num(settings.get(self.window_title, "width", "-1"))
                height = num(settings.get(self.window_title, "height", "-1"))

            if -1 in [width, height]:  # bad settings or first run
                if self.mode != "grid" and sum(self.q2_form.init_size) > 0:
                    # init size given
                    width, height = self.get_init_size()
                else:
                    width = paw.parent().size().width() * 0.9 if self.mode == "grid" else 0.5
                    height = paw.parent().size().height() * 0.9 if self.mode == "grid" else 0.5

            paw.resize(int(width), int(height))

            sizeAfter = paw.size()
            self.expand_size(paw, sizeBefore, sizeAfter)

            if self.q2_form.do_not_save_geometry:
                left, top = self.center_pos()
            else:
                left = num(settings.get(self.window_title, "left", "-1"))
                top = num(settings.get(self.window_title, "top", "-1"))
                if -1 in [left, top]:  # bad settings or first run
                    left, top = self.center_pos()

            paw.move(int(left), int(top))

            self.fit_size_and_pos(paw)

            if not self.q2_form.do_not_save_geometry:
                if num(settings.get(self.window_title, "is_max", "0")):
                    self.showMaximized()
                    paw.move(0, 0)

    def center_pos(self):
        left = int((self.parent().parent().size().width() - self.parent().size().width()) / 2)
        top = int((self.parent().parent().size().height() - self.parent().size().height()) / 2)
        return left, top

    def expand_size(self, paw, sizeBefore, sizeAfter):
        wDelta = 0 if sizeBefore.width() < sizeAfter.width() else sizeBefore.width() - sizeAfter.width()
        hDelta = 0 if sizeBefore.height() < sizeAfter.height() else sizeBefore.height() - sizeAfter.height()
        if wDelta or hDelta:
            paw.resize(paw.size().width() + wDelta, paw.size().height() + hDelta)

    def get_init_size(self):
        w, h = self.q2_form.init_size
        parent_size = self.parent().parent().size()
        return [int(parent_size.width() * w / 100), int(parent_size.height() * h / 100)]

    def fit_size_and_pos(self, paw):
        """ensure form fits outside window"""
        parent_size = paw.parent().size()

        size = paw.size()
        original_size = paw.size()

        pos = paw.pos()
        orginal_pos = paw.pos()
        # width
        if parent_size.width() - (size.width()) < 0:
            size.setWidth(parent_size.width())
        if pos.x() + size.width() > parent_size.width():
            pos.setX(parent_size.width() - size.width())
        if pos.x() < 0:
            pos.setX(0)
        # height
        if parent_size.height() - (size.height()) < 0:
            size.setHeight(parent_size.height())

        if pos.y() + size.height() > parent_size.height():
            pos.setY(parent_size.height() - size.height())
        if pos.y() < 0:
            pos.setY(0)
        if orginal_pos != pos:
            paw.move(pos)
        if size != original_size:
            paw.resize(size)

    def set_position(self, left, top):
        left = int(left)
        top = int(top)
        paw = self.parent()
        if paw is not None:
            paw.move(left, top)
        else:
            self.move(left, top)

    def set_size(self, w, h):
        w = int(w)
        h = int(h)
        paw = self.parent()
        if paw is not None:
            paw.resize(w, h)
        else:
            self.resize(w, h)

    def get_position(self):
        parent_mdi_sub_window = self.parent()
        if parent_mdi_sub_window is not None:
            return (parent_mdi_sub_window.pos().x(), parent_mdi_sub_window.pos().y())

    def showEvent(self, event=None):
        if self.shown:
            return
        if self not in self.q2_form.form_stack:
            self.q2_form.form_stack.append(self)

        if event:
            event.accept()

        self.q2_form.form_is_active = True
        self.shown = True
        if hasattr(q2app.q2_app, "settings"):
            self.restore_geometry(q2app.q2_app.settings)

        if hasattr(self.parent(), "setWindowFlag"):
            self.parent().setWindowFlag(Qt.WindowType.WindowMinimizeButtonHint, False)

            if self.q2_form.hide_title:
                self.parent().setWindowFlags(
                    Qt.WindowType.CustomizeWindowHint
                    | Qt.WindowType.FramelessWindowHint
                    | Qt.WindowType.WindowStaysOnBottomHint
                    | Qt.WindowType.WindowCloseButtonHint
                )
        if self.q2_form.maximized:
            self.showMaximized()

        self.activateWindow()

        if not isinstance(self.parent(), QMdiSubWindow):
            self.escape_enabled = False
        else:
            self.parent().windowStateChanged.connect(self.mdi_subwindow_state_changed)
            self.mdi_subwindow_state_changed()

        if hasattr(self, "on_activate"):
            self.on_activate()

        if self.mode == "form":
            self.q2_form.after_form_show()
            self.q2_form.show_()
        elif self.mode == "grid":
            self.q2_form.after_grid_show()

    def mdi_subwindow_state_changed(self):
        if self.about_to_close:
            return
        if self.isMaximized():
            self.parent().setWindowFlag(Qt.WindowType.FramelessWindowHint, True)
            q2app.q2_app.q2_tabwidget.show_mdi_normal_button(True)
        else:
            self.parent().setWindowFlag(Qt.WindowType.FramelessWindowHint, False)
            self.parent().setWindowFlag(Qt.WindowType.WindowMinimizeButtonHint, False)
            q2app.q2_app.q2_tabwidget.show_mdi_normal_button(False)

    def keyPressEvent(self, event: QEvent):
        key = event.key()
        try:
            keyText = QKeySequence(key).toString()
        except Exception:
            keyText = ""
        if self.mode == "form" and key in (Qt.Key.Key_End,):
            keyText = "PgDown"
        if key == Qt.Key.Key_Escape and self.escape_enabled:
            self.close()
        elif key == Qt.Key.Key_Escape and not self.escape_enabled:
            event.ignore()
            # return
        elif self.mode == "form" and key in (Qt.Key.Key_Up,):
            QApplication.sendEvent(
                self, QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Tab, Qt.KeyboardModifier.ShiftModifier)
            )
        elif self.mode == "form" and key in (Qt.Key.Key_Enter, Qt.Key.Key_Return, Qt.Key.Key_Down):
            QApplication.sendEvent(self, QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Tab, event.modifiers()))
        elif self.mode == "grid" and key in (Qt.Key.Key_Return,):
            self.q2_form.grid_double_clicked()
        elif keyText in self.hotkey_widgets:  # is it form hotkey
            for widget in self.hotkey_widgets[keyText]:
                if widget.is_enabled() and hasattr(widget, "valid"):
                    widget.valid()
                    return

    def close(self):
        super().close()
        if self.parent() is not None:
            if isinstance(self.parent(), QMdiSubWindow):
                self.parent().close()
                # QDialog.close(self)
        else:
            QDialog.close(self)

    def closeEvent(self, event=None):
        super().close()
        self.q2_form._close()
        self.not_closed = False
        self.q2_form.form_is_active = False
        if self.heap.prev_mdi_window:
            self.heap.prev_mdi_window.setEnabled(True)
            # else:
            #     q2app.q2_app.q2_tabwidget.show_mdi_normal_button(False)
            if self.heap.prev_focus_widget is not None and not isinstance(
                self.heap.prev_focus_widget, QTabBar
            ):
                try:
                    self.heap.prev_focus_widget.setFocus()
                except Exception:
                    pass
            elif hasattr(self.heap.prev_mdi_window, "setFocus"):
                self.heap.prev_mdi_window.setFocus()
        else:
            q2app.q2_app.q2_tabwidget.show_mdi_normal_button(False)
        q2app.q2_app.set_tabbar_text(self.heap.prev_tabbar_text)
        self.about_to_close = True
        if event:
            event.accept()

    def set_style_sheet(self, css):
        self.setStyleSheet(css)


# Tells the module which engine to use
q2gui.q2dialogs.Q2Form = Q2Form
# q2Mess
# q2Wait
# q2AskYN



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import QTableWidget, QSizePolicy, QStyle
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPalette


from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.pyqt6.widgets.q2label import q2label

from q2gui.q2utils import int_


class q2sheet(QTableWidget, Q2Widget):
    def __init__(self, meta):
        # super().__init__({"label": meta.get("label", "")})
        super().__init__(meta)
        self.column_headers = []
        self.row_headers = []
        # print()
        self.selection_background_color = "yellow"
        self.selection_background_color = self.palette().color(QPalette.ColorRole.Highlight).toRgb().name()
        self.horizontalHeader().setMinimumSectionSize(0)
        self.verticalHeader().setMinimumSectionSize(0)
        self.auto_expand = False
        self.setEditTriggers(self.EditTrigger.NoEditTriggers)
        self.spaned_cells = []

        self.sheet_styles = {}
        self.cell_styles = {}

        if self.meta.get("when"):
            self.itemSelectionChanged.connect(self.meta.get("when"))
        if self.meta.get("valid"):
            self.itemSelectionChanged.connect(self.meta.get("valid"))

    def mousePressEvent(self, event):
        rez = super().mousePressEvent(event)
        if self.meta.get("when"):
            self.meta.get("when")()
        return rez

    def focusInEvent(self, event):
        rez = super().focusInEvent(event)
        if event.reason() not in (Qt.FocusReason.PopupFocusReason,):
            x = self.currentIndex()
            if x.isValid():
                self.setCurrentCell(x.row(), x.column())
            else:
                self.setCurrentCell(0, 0)
        if self.meta.get("when"):
            self.meta.get("when")()
        return rez

    def focusOutEvent(self, event):
        if self.meta.get("valid"):
            if self.meta.get("valid")() is False:
                self.setFocus()
                return
        rez = super().focusOutEvent(event)
        return rez

    def mouseDoubleClickEvent(self, e):
        if self.meta.get("form"):
            self.meta.get("form").grid_double_clicked()
        return super().mouseDoubleClickEvent(e)

    def keyPressEvent(self, ev):
        if ev.key() in [Qt.Key.Key_Enter, Qt.Key.Key_Return] and self.meta.get("eat_enter"):
            if self.meta.get("form"):
                if self.meta.get("dblclick"):
                    self.meta.get("dblclick")()
                else:
                    self.meta.get("form").grid_double_clicked()
        else:
            super().keyPressEvent(ev)

    def set_auto_expand(self, mode=True):
        self.auto_expand = mode

    def expand(self):
        if self.auto_expand:
            self.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum))
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

            frame_delta = self.style().pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth) * 0

            height = self.horizontalHeader().height()
            for x in range(self.rowCount()):
                height += self.rowHeight(x)
            self.setFixedHeight(height + frame_delta)

            width = self.verticalHeader().width()
            for x in range(self.columnCount()):
                width += self.columnWidth(x)
            self.setFixedWidth(width + frame_delta)
        else:
            self.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
            self.setMinimumWidth(0)
            self.setMinimumHeight(0)

    def set_row_count(self, row=2):
        self.setRowCount(row)
        self.expand()

    def set_column_count(self, column=2):
        self.setColumnCount(column)
        self.expand()

    def set_row_column_count(self, row=2, column=2):
        self.set_row_count(row)
        self.set_column_count(column)
        self.expand()

    def set_row_header_size(self, size):
        self.verticalHeader().setFixedWidth(size)
        self.expand()

    def set_column_headers(self, headers_list=[]):
        self.column_headers = headers_list
        self.setHorizontalHeaderLabels(self.column_headers)
        self.expand()

    def set_row_headers(self, headers_list=[]):
        self.row_headers = headers_list
        self.setVerticalHeaderLabels(self.row_headers)
        self.expand()

    def set_column_header(self, column=0, header=""):
        if column >= self.columnCount():
            return

        if not column < len(self.column_headers):
            for x in range(len(self.column_headers), self.columnCount()):
                self.column_headers.append("")
        self.column_headers[column] = header
        self.set_column_headers(self.column_headers)
        self.expand()

    def set_row_header(self, row=0, header=""):
        if row >= self.rowCount():
            return
        if not row < len(self.row_headers):
            for x in range(len(self.row_headers) - 1, self.rowCount()):
                self.row_headers.append("")
        self.row_headers[row] = header
        self.set_row_headers(self.row_headers)
        self.expand()

    def hide_column_headers(self):
        self.horizontalHeader().hide()
        self.expand()

    def show_column_headers(self):
        self.horizontalHeader().show()
        self.expand()

    def hide_row_headers(self):
        self.verticalHeader().hide()
        self.expand()

    def show_row_headers(self):
        self.verticalHeader().show()
        self.expand()

    def set_column_size(self, width=[], column=None):
        if isinstance(width, list):
            for column, size in enumerate(width):
                self.set_column_size(size, column)
        elif isinstance(width, int) and column is None:
            for x in range(self.columnCount()):
                self.set_column_size(width, x)
        else:
            self.setColumnWidth(column, width)
        self.expand()

    def get_row_size(self, row=None):
        return self.rowHeight(row)

    def set_row_size(self, heights=[], row=None):
        if isinstance(heights, list):
            for row, size in enumerate(heights):
                self.set_row_size(size, row)
        elif isinstance(heights, int) and row is None:
            for x in range(self.rowCount()):
                self.set_row_size(heights, x)
        else:
            self.setRowHeight(row, int(heights))
        self.expand()

    def clear_selection(self):
        self.clearSelection()

    def get_selection(self):
        return [(x.row(), x.column()) for x in self.selectedIndexes()]

    def clear_spans(self):
        self.spaned_cells = []
        self.clearSpans()
        self.updateGeometries()

    def set_span(self, row, column, row_span, column_span):
        self.setSpan(row, column, row_span, column_span)
        for sr in range(row, row + row_span):
            for sc in range(column, column + column_span):
                if sr != row or sc != column:
                    self.spaned_cells.append((sr, sc))
                    # self.get_cell_widget(sr, sc).setVisible(False)
                    self.removeCellWidget(sr, sc)
        self.expand()
        self.updateGeometries()

    def get_cell_text(self, row=None, column=None):
        rez = []
        if row is None and column is None:
            row = 0
        if row is not None and column is None:
            for column in range(self.columnCount()):
                rez.append(self.get_cell_text(row, column))
        elif row is None and column is not None:
            for row in range(self.rowCount()):
                rez.append(self.get_cell_text(row, column))
        else:
            row = int_(row)
            column = int_(column)
            rez = self.get_cell_widget(row, column).get_text()
        return rez

    def get_cell_style_sheet(self, row=None, column=None):
        rez = []
        if row is None and column is None:
            row = 0
        if row is not None and column is None:
            for column in range(self.columnCount()):
                rez.append(self.get_cell_style(row, column))
        elif row is None and column is not None:
            for row in range(self.rowCount()):
                rez.append(self.get_cell_style(row, column))
        else:
            row = int_(row)
            column = int_(column)
            rez = self.get_cell_widget(row, column).get_style_sheet()
        return rez

    def get_text(self):
        return self.get_cell_widget(self.current_row(), self.current_column()).get_text()

    def set_text(self, text):
        self.set_cell_text(text, self.current_row(), self.current_column())

    def set_cell_text(self, text="", row=None, column=None):
        if isinstance(text, list):
            if row is None and column is None:
                row = 0
            if row is not None:
                for x in range(self.columnCount()):
                    if x < len(text):
                        self.set_cell_text(text[x], row, x)
            else:
                for x in range(self.rowCount()):
                    if x < len(text):
                        self.set_cell_text(text[x], x, column)
        else:
            row = int_(row)
            column = int_(column)
            cell_widget = self.get_cell_widget(row, column)
            cell_widget.setText(text)

    def set_cell_style_sheet(self, style_text="", row=None, column=None):
        if row is None and column is None:
            row = 0
        if row is not None and column is None:
            for x in range(self.columnCount()):
                if isinstance(style_text, dict):
                    if x < len(style_text):
                        self.set_cell_style_sheet(style_text[x], row, x)
                else:
                    self.set_cell_style_sheet(style_text, row, x)
        elif row is None and column is not None:
            for x in range(self.rowCount()):
                if isinstance(style_text, dict):
                    if x < len(style_text):
                        self.set_cell_style_sheet(style_text[x], x, column)
                else:
                    self.set_cell_style_sheet(style_text, x, column)
        else:  # row and column - set style
            row = int_(row)
            column = int_(column)
            cell_widget = self.get_cell_widget(row, column)
            if (row, column) in self.spaned_cells:
                cell_widget.set_style_sheet("")
                cell_widget.setVisible(False)
                # print(row, column)
                return
            # else:
            #     cell_widget.setVisible(True)
            cell_key = f"{row},{column}"
            if style_text is None and cell_key not in self.cell_styles:
                style_text = cell_widget.get_style_sheet()
            if (row, column) in [(x.row(), x.column()) for x in self.selectedIndexes()]:
                if isinstance(style_text, dict):
                    style_text["background-color"] = self.selection_background_color
                elif isinstance(style_text, str):
                    style_text += f";background-color:{self.selection_background_color};"
                else:  # None - using self.cell_styles
                    style_text = dict(self.sheet_styles)
                    style_text.update(self.cell_styles.get(cell_key, {}))
                    style_text["background-color"] = self.selection_background_color
            else:
                if isinstance(style_text, str):
                    style_text = style_text.replace(f"background-color:{self.selection_background_color}", "")
                elif style_text is None:
                    style_text = dict(self.sheet_styles)
                    style_text.update(self.cell_styles.get(cell_key, {}))

            cell_widget.set_style_sheet(style_text)

    def selectionChanged(self, selected, deselected):
        rez = super().selectionChanged(selected, deselected)
        for x in deselected.indexes():
            # st = self.get_cell_widget(x.row(), x.column()).get_style_sheet()
            self.set_cell_style_sheet(None, x.row(), x.column())

        for x in selected.indexes():
            # st = self.get_cell_widget(x.row(), x.column()).get_style_sheet()
            self.set_cell_style_sheet(None, x.row(), x.column())
        return rez

    def get_cell_widget(self, row, column):
        cell_widget = self.cellWidget(row, column)
        if cell_widget is None:
            cell_widget = q2label({"label": "", "dblclick": self.meta.get("dblclick")})
            cell_widget.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
            cell_widget.set_maximum_height(9999)
            self.setCellWidget(row, column, cell_widget)
        return cell_widget

    def get_current_widget(self):
        return self.get_cell_widget(self.currentRow(), self.currentColumn())

    def current_row(self):
        return self.currentRow()

    def current_column(self):
        return self.currentColumn()

    def set_current_cell(self, row, column):
        self.setCurrentCell(row, column)

    def insert_row(self, after_row=None):
        self.insertRow(after_row)
        # self.setRowCount(self.rowCount() + 1)
        # for row in range(self.rowCount() - 1, after_row, -1):
        #     for col in range(self.columnCount()):
        #         self.set_cell_text(self.get_cell_text(row - 1, col), row, col)
        #         self.set_cell_style_sheet(self.get_cell_style_sheet(row - 1, col), row, col)
        #         if row == after_row + 1:
        #             self.set_cell_text("", row - 1, col)
        #             self.set_cell_style_sheet("", row - 1, col)

    def move_row(self, after_row):
        for col in range(self.columnCount()):
            text = self.get_cell_text(after_row, col)
            style = self.get_cell_style_sheet(after_row, col)
            self.set_cell_text(self.get_cell_text(after_row + 1, col), after_row, col)
            self.set_cell_style_sheet(self.get_cell_style_sheet(after_row + 1, col), after_row, col)
            self.set_cell_text(text, after_row + 1, col)
            self.set_cell_style_sheet(style, after_row + 1, col)

    def remove_row(self, remove_row):
        self.removeRow(remove_row)
        # for row in range(remove_row, self.rowCount()):
        #     for col in range(self.columnCount()):
        #         self.set_cell_text(self.get_cell_text(row + 1, col), row, col)
        #         self.set_cell_style_sheet(self.get_cell_style_sheet(row + 1, col), row, col)
        # self.setRowCount(self.rowCount() - 1)

    def insert_column(self, after_column=None):
        self.insertColumn(after_column)
        # self.setColumnCount(self.columnCount() + 1)
        # for col in range(self.columnCount() - 1, after_column, -1):
        #     for row in range(self.rowCount()):
        #         self.set_cell_text(self.get_cell_text(row, col - 1), row, col)
        #         self.set_cell_style_sheet(self.get_cell_style_sheet(row, col - 1), row, col)
        #         if col == after_column + 1:
        #             self.set_cell_text("", row, col - 1)
        #             self.set_cell_style_sheet("", row, col - 1)

    def remove_column(self, remove_column):
        self.removeColumn(remove_column)
        # for col in range(remove_column, self.columnCount()):
        #     for row in range(self.rowCount()):
        #         self.set_cell_text(self.get_cell_text(row, col + 1), row, col)
        #         self.set_cell_style_sheet(self.get_cell_style_sheet(row, col + 1), row, col)
        # self.setColumnCount(self.columnCount() - 1)

    def move_column(self, after_column):
        for row in range(self.rowCount()):
            text = self.get_cell_text(row, after_column)
            style = self.get_cell_style_sheet(row, after_column)
            self.set_cell_text(self.get_cell_text(row, after_column + 1), row, after_column)
            self.set_cell_style_sheet(self.get_cell_style_sheet(row, after_column + 1), row, after_column)
            self.set_cell_text(text, row, after_column + 1)
            self.set_cell_style_sheet(style, row, after_column + 1)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtGui import QMouseEvent, QColor, QPalette
from PyQt6.QtWidgets import QFrame, QColorDialog
from PyQt6.QtCore import Qt

import q2gui.q2app as q2app
from q2gui.q2form import Q2Form
from q2gui.q2utils import num
from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.pyqt6.q2window import Q2Frame
from q2gui.pyqt6.widgets.q2line import q2line
from q2gui.pyqt6.widgets.q2button import q2button


class q2color(QFrame, Q2Widget, Q2Frame):
    def __init__(self, meta):
        super().__init__(meta)
        Q2Frame.__init__(self, "h")
        self.meta = meta
        self._color = QColor(meta.get("data", "black"))

        # --- Button ---
        self.btn = q2button(
            {
                "label": "",
                "valid": self.choose_color,
                "form": self.meta["form"],
            }
        )
        self.btn.setFixedSize(20, 20)
        # --- Line ---
        self.meta_changed = self.meta.get("changed")
        self.meta["changed"] = self.line_changed
        self.line = q2line(meta)

        # --- Layout ---
        self.add_widget(self.btn)
        self.add_widget(self.line)
        self.set_content_margins(0)
        self.layout().setSpacing(3)
        self.setFocusProxy(self.line)
        self._update_button_color()

    # --- Logic ---
    def _update_button_color(self):
        self.btn.set_style_sheet(f"background-color:{self.line.get_text()}")

    def choose_color(self):
        new_color = QColorDialog.getColor(QColor(self.line.get_text()), self, "Select Color")
        if new_color.isValid():
            self.set_color(new_color)

    def line_changed(self, text):
        if self.meta_changed:
            self.meta_changed()
        color = QColor(text.strip())
        if color.isValid():
            self._update_button_color()

    # --- Public API ---
    def set_color(self, color):
        if hasattr(self, "line"):
            self.line.set_text(color.name())

    def set_text(self, text):
        color = QColor(text)
        if color.isValid():
            self.set_color(color)

    def text(self):
        return self.line.get_text()



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import QGroupBox, QSplitter
from PyQt6.QtCore import Qt
from PyQt6.QtCore import QSize


from q2gui.pyqt6.q2window import Q2Frame
from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.q2utils import int_


class q2frame(QGroupBox, Q2Widget, Q2Frame):
    def __init__(self, meta):
        super().__init__(meta)
        Q2Frame.__init__(self, meta.get("column", "/v")[1])
        self.meta = meta
        self.splitter = None
        self.scroller = None
        if meta.get("column", "")[2:3] == "s":  # Splitter!
            self.splitter = q2splitter()
            if meta.get("column").startswith("/v"):
                self.splitter.setOrientation(Qt.Orientation.Vertical)
            self.layout().addWidget(self.splitter)
        if meta.get("label") not in ("", "-") and not meta.get("check"):
            self.set_title(meta.get("label"))
            self.setObjectName("title")
        if meta.get("label", "") == "":
            self.hide_border()
        elif meta.get("label", "") == "-":
            self.setStyleSheet("QGroupBox {border:1px solid lightgray}")

    def hide_border(self):
        self.setObjectName("grb")
        self.set_title("")
        self.add_style_sheet(" QGroupBox#grb {border:0} ")

    def set_title(self, title):
        self.setTitle(title)

    def can_get_focus(self):
        return False

    def get_widget_count(self):
        return self.layout().count()

    def add_widget(self, widget=None, label=None):
        if self.splitter is not None:
            self.splitter.addWidget(widget)
            if hasattr(widget, "meta"):
                self.splitter.setStretchFactor(self.splitter.count() - 1, int_(widget.meta.get("stretch", 0)))
        else:
            return super().add_widget(widget=widget, label=label)


class q2splitter(QSplitter):
    def __init__(self):
        super().__init__()
        # self.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))

    def get_sizes(self):
        return ",".join([f"{x}" for x in self.sizes()])

    def set_sizes(self, sizes):
        if sizes == "":
            init_sizes = [int_(self.widget(x).meta.get("stretch", 1)) for x in range(self.count())]
            init_sizes = [x if x > 0 else 1 for x in init_sizes]
            if sum(init_sizes):
                widget_size = (
                    self.width() if self.orientation() is Qt.Orientation.Horizontal else self.height()
                )
                init_sizes = [str(int(x * widget_size / sum(init_sizes))) for x in init_sizes]
                for x in range(self.count()):
                    widgget = self.widget(x)
                    if widgget.meta.get("control") == "toolbar":
                        init_sizes[x] = str(widgget.sizeHint().height())
                sizes = ",".join(init_sizes)
        if sizes:
            sizes = [int(x) for x in sizes.split(",")]
            self.setSizes(sizes)

    def showEvent(self, ev):
        self.updateGeometry()
        return super().showEvent(ev)

    def sizeHint(self):
        if self.isVisible():
            return QSize(99999, 99999)
        else:
            return super().sizeHint()



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os


import codecs

from PyQt6.QtWidgets import QLabel, QApplication, QSizePolicy
from PyQt6.QtGui import QPixmap, QImage
from PyQt6.QtCore import QByteArray, QBuffer, QIODevice, QFile, Qt

from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.q2app import Q2Actions
import q2gui.q2app as q2app


class q2image(QLabel, Q2Widget):
    def __init__(self, meta={}):
        actions = Q2Actions()
        # actions.show_main_button = 0
        # actions.show_actions = 0
        actions.add_action("Load", self.load_image_from_file)
        actions.add_action("Save", self.save_image_to_file)
        actions.add_action("-")
        actions.add_action("Paste", self.clipboard_paste)
        actions.add_action("Copy", self.clipboard_copy)
        actions.add_action("-")
        actions.add_action("Clear", self.clear_image)
        meta["actions"] = actions
        super().__init__(meta)
        self.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Ignored)

        self.image_base64 = ""
        self.image = None

        self.set_text(self.meta["data"])
        self.set_style_sheet("{border: 1px solid black; border-radius:0px; margin:0px; padding:0px }")

    def clear_image(self):
        self.set_qimage(image=QImage())

    def clipboard_copy(self):
        clipboard = QApplication.clipboard()
        clipboard.setPixmap(QPixmap.fromImage(self.image))

    def clipboard_paste(self):
        clipboard = QApplication.clipboard()
        if not clipboard.image(clipboard.Mode.Clipboard).isNull():
            self.set_qimage(image=clipboard.image(clipboard.Mode.Clipboard))
        elif clipboard.mimeData().hasUrls():
            filename = clipboard.mimeData().urls()[0].toLocalFile()
            if "LNK" == filename[-3:].upper():
                filename = QFile.symLinkTarget(filename)
            try:
                self.set_qimageimage = QImage(filename)
            except Exception:
                pass

    def save_image_to_file(self):
        image_file, image_type = q2app.q2_app.get_save_file_dialoq(
            "Save image",
            filter="PNG (*.png);;JPG(*.jpg)",
        )
        if image_file:
            self.image.save(image_file)

    def load_image_from_file(self, image_file=""):
        if not os.path.isfile(image_file):
            image_file = q2app.q2_app.get_open_file_dialoq(
                "Load image",
                filter="Images (*.png *.jpg)",
            )[0]
        if image_file:
            self.set_qimage(image=QImage(image_file))

    def get_base64(self):
        ba = QByteArray()
        buffer = QBuffer(ba)
        buffer.open(QIODevice.OpenModeFlag.WriteOnly)
        self.image.save(buffer, "PNG")
        return bytes(ba.toBase64()).decode("utf-8")

    def ensure_base64(self, text):
        try:
            text = text.replace("\n", "").replace(" ", "")
            int(text, 16)
            text = codecs.encode(codecs.decode(text, "hex"), "base64").decode()
        except Exception:
            pass
        finally:
            return text

    def resizeEvent(self, ev):
        rez = super().resizeEvent(ev)
        self.keep_size()
        return rez

    def set_qimage(self, image: QImage = None, image_base64=None):
        if image_base64:
            self.image_base64 = image_base64
            self.image = QImage.fromData(QByteArray.fromBase64(bytes(self.image_base64, "utf8")))
        elif image:
            self.image = image
            self.image_base64 = self.get_base64()
        else:
            self.image = QImage()
            self.image_base64 = None

        self.setPixmap(QPixmap.fromImage(self.image))
        self.keep_size()

    def keep_size(self):
        if (
            self.size().width() < self.image.size().width()
            or self.size().height() < self.image.size().height()
        ):
            self.setPixmap(
                QPixmap.fromImage(self.image).scaled(self.size(), Qt.AspectRatioMode.KeepAspectRatio)
            )

    def set_text(self, text):
        self.set_qimage(image_base64=self.ensure_base64(text))

    def get_text(self):
        return self.image_base64
